#pragma once
#include <dios/sys/task.hpp>
#include <dios/sys/options.hpp>
#include <dios/sys/syscall.hpp>
#include <dios/sys/sched_null.hpp>
#include <sys/mman.h>

extern "C" void __dios_native_exit( int );
extern "C" void *__dios_native_mmap( void *, size_t, int, int, int, off_t );
extern "C" void __dios_native_puts( const char * );

namespace __dios
{
    extern jmp_buf sched_context;
    extern const sighandler_t defhandlers[ 32 ];

    static void *stack_sink;

    template < typename next >
    struct sched_native : public sched_null< next >
    {
        using typename sched_null< next >::task;
        using tasks = task_array< task >;
        using typename sched_null< next >::process;

        template< typename Setup >
        void setup( Setup s )
        {
            sched_null< next >::setup( s );
            start_task( main_entry, this->_task._arg, 0 );
            __vm_ctl_set( _VM_CR_Scheduler,
                          reinterpret_cast< void * >( run_scheduler< typename Setup::Context > ) );
        }

        task *choose_task() noexcept
        {
            if ( _tasks.empty() )
                return nullptr;
            return _tasks[ __vm_choose( _tasks.size() ) ].get();
        }
#if 0
        template< typename I >
        void erase_processes( I pivot )
        {
            ArrayMap< Process *, bool > p;
            for ( auto i = pivot; i != tasks.end(); ++i )
                p.emplace( (*i)->_proc, true );
            for ( auto i = tasks.begin(); i != pivot; ++i )
                p.erase( (*i)->_proc );
            for ( auto item : p )
                delete item.first;
        }

        void kill_process( pid_t id ) noexcept
        {
            if ( !id )
            {
                size_t c = tasks.size();
                erase_processes( tasks.begin() );
                tasks.erase( tasks.begin(), tasks.end() );
                this->reschedule();
            }

            bool resched = false;
            auto r = std::partition( tasks.begin(), tasks.end(), [&]( auto& t )
                                     {
                                         if ( t->_proc->pid != id )
                                             return true;
                                         if ( t->_tls == __dios_this_task() )
                                             resched = true;
                                         return false;
                                     } );

            erase_processes( r );
            tasks.erase( r, tasks.end() );
            if ( resched )
                this->reschedule();
        }
#endif
        task *find_task( __dios_task id ) { return _tasks.find( id ); }
        task *current_task()
        {
            __dios_trace_f( "this_task = %p, current_task = %p", __dios_this_task(),
                            _tasks.find( __dios_this_task() ) );
            return _tasks.find( __dios_this_task() );
        }

        process *current_process() { return current_task()->_proc; }
        pid_t getpid() { return 1; }

        static void main_entry( void *arg )
        {
            auto f = static_cast< main_args * >( arg );
            __dios_start( f->l, f->argc, f->argv, f->envp );
        }

        [[noinline]] static void trampoline( jmp_buf *ctx, void (*f)( void * ), void *arg )
        {
            if ( !setjmp( *ctx ) )
                return;
            f( arg );
        }

        static void mkstack( jmp_buf *ctx, void (*f)( void * ), void *arg )
        {
            int size = 800000;
            void *map = __dios_native_mmap( 0, size, PROT_READ | PROT_WRITE,
                                            MAP_ANON | MAP_STACK | MAP_PRIVATE, -1, 0 );
            char *stack = static_cast< char * >( map ) + size;
            if ( map == MAP_FAILED )
                __dios_native_exit( 100 );
            stack_sink = __builtin_alloca( reinterpret_cast< char * >( &stack ) - stack );
            trampoline( ctx, f, arg );
        }

        __dios_task start_task( __dios_task_routine routine, void *arg, int tls_size )
        {
            auto t = _tasks.emplace_back( new task ).get();
            t->_arg = arg;
            t->_proc = this->_task._proc;
            mkstack( &t->_context, routine, arg );
            return t->get_id();
        }
/*
        void exit_process( int code )
        {
            kill_process( 0 );
        }
*/
        void kill_task( __dios_task id )
        {
            bool res = _tasks.remove( id );
            __dios_assert_v( res, "Killing non-existing task" );
            if ( id == __dios_this_task() )
                this->reschedule();
        }

        __dios_task *get_process_tasks( __dios_task tid )
        {
            auto ret = static_cast< __dios_task * >( __vm_obj_make( sizeof( __dios_task ) * _tasks.size(),
                                                                    _VM_PT_Heap ) );
            int i = 0;
            for ( auto &t : _tasks )
                ret[ i++ ] = t->_tls;

            return ret;
        }

        bool check_final()
        {
            return _tasks.empty();
        }

        __inline void prepare( task &t )
        {
            __vm_ctl_set( _VM_CR_User1, &t._context );
            __vm_ctl_set( _VM_CR_User2, t.get_id() );
            __vm_ctl_set( _VM_CR_User3, this->debug );
        }

        __inline void monitor( task &t )
        {
            this->runMonitors();
        }

        __inline void jump( task &t )
        {
            __vm_ctl_flag( _VM_CF_KernelMode | _VM_CF_IgnoreCrit | _VM_CF_IgnoreLoop, 0 );
            // __dios_trace_f( "sched_native::jump( %p )", &t );

            if ( !setjmp( sched_context ) )
                longjmp( t._context, 1 );
            __vm_ctl_flag( 0, _VM_CF_KernelMode | _VM_CF_IgnoreCrit | _VM_CF_IgnoreLoop );
        }

        __inline void run( task &t )
        {
            prepare( t );
            monitor( t ); /* hmm. */
            jump( t );
        }

        template < typename Context >
        static void run_scheduler() noexcept
        {
            auto& scheduler = get_state< Context >();
            task *t;

            while ( t = scheduler.choose_task() )
                scheduler.run( *t );
        }

        tasks _tasks;
    };
}
